from ._typing import SUPPRESS, Action, Annotated, FileProtocol, TypedFileType
from .argspec import ArgumentSpec, ArgumentSpecType
from .arguments import RunnableArguments, SubcommandArguments
from .base import BaseArguments
from .subcommand import SubcommandSpec, subcommand

__all__ = [
    # Core classes
    "BaseArguments",
    "ArgumentSpec",
    # Subcommand support (recommended)
    "subcommand",
    "SubcommandSpec",
    # Advanced features
    "RunnableArguments",
    "SubcommandArguments",
    "ArgumentSpecType",
    # Utilities
    "SUPPRESS",
    "Action",
    "FileProtocol",
    "TypedFileType",
    "Annotated",
]



import unittest
from typing import Optional

from spargear import ArgumentSpec, BaseArguments, subcommand


class GitCommitArguments(BaseArguments):
    """Git commit command arguments."""

    message: ArgumentSpec[str] = ArgumentSpec(["-m", "--message"], required=True, help="Commit message")
    amend: ArgumentSpec[bool] = ArgumentSpec(["--amend"], action="store_true", help="Amend previous commit")


class GitPushArguments(BaseArguments):
    """Git push command arguments."""

    remote: ArgumentSpec[str] = ArgumentSpec(["remote"], nargs="?", default="origin", help="Remote name")
    branch: ArgumentSpec[Optional[str]] = ArgumentSpec(["branch"], nargs="?", help="Branch name")
    force: ArgumentSpec[bool] = ArgumentSpec(["-f", "--force"], action="store_true", help="Force push")


class GitStatusArguments(BaseArguments):
    """Git status command arguments."""

    short: ArgumentSpec[bool] = ArgumentSpec(["-s", "--short"], action="store_true", help="Show short format")


class GitArgumentsWithDecorator(BaseArguments):
    """Git command line interface using @subcommand decorator."""

    verbose: ArgumentSpec[bool] = ArgumentSpec(["-v", "--verbose"], action="store_true", help="Increase verbosity")

    @subcommand(help="Record changes to the repository")
    def commit():
        """Record changes to the repository.

        This command records the changes in the repository.
        """
        return GitCommitArguments

    @subcommand(name="push", help="Update remote refs")
    def push_cmd():
        return GitPushArguments

    @subcommand()
    def status():
        """Show the working tree status.

        Displays paths that have differences between the index file
        and the current HEAD commit.
        """
        return GitStatusArguments


class TestSubcommandDecorator(unittest.TestCase):
    def test_decorator_with_help(self):
        """Test that @subcommand with help parameter works."""
        args = GitArgumentsWithDecorator(["commit", "-m", "test commit"])
        commit = args.last_subcommand
        assert isinstance(commit, GitCommitArguments), "commit should be a GitCommitArguments instance"
        self.assertEqual(commit.message.unwrap(), "test commit")
        self.assertFalse(commit.amend.unwrap())

    def test_decorator_with_name_override(self):
        """Test that @subcommand with custom name works."""
        args = GitArgumentsWithDecorator(["push", "upstream", "main", "--force"])
        push = args.last_subcommand
        assert isinstance(push, GitPushArguments), "push should be a GitPushArguments instance"
        self.assertEqual(push.remote.unwrap(), "upstream")
        self.assertEqual(push.branch.unwrap(), "main")
        self.assertTrue(push.force.unwrap())

    def test_decorator_with_docstring(self):
        """Test that @subcommand extracts help from docstring."""
        args = GitArgumentsWithDecorator(["status", "--short"])
        status = args.last_subcommand
        assert isinstance(status, GitStatusArguments), "status should be a GitStatusArguments instance"
        self.assertTrue(status.short.unwrap())

    def test_subcommand_specs_created(self):
        """Test that SubcommandSpec instances are properly created."""
        cls = GitArgumentsWithDecorator

        # Check that subcommands are registered
        self.assertIn("commit", cls.__subcommands__)
        self.assertIn("push", cls.__subcommands__)
        self.assertIn("status", cls.__subcommands__)

        # Check commit subcommand spec
        commit_spec = cls.__subcommands__["commit"]
        self.assertEqual(commit_spec.name, "commit")
        self.assertEqual(commit_spec.help, "Record changes to the repository")
        self.assertIsNotNone(commit_spec.argument_class_factory)
        self.assertIsNone(commit_spec.argument_class)
        self.assertEqual(commit_spec.get_argument_class(), GitCommitArguments)

        # Check push subcommand spec (with custom name)
        push_spec = cls.__subcommands__["push"]
        self.assertEqual(push_spec.name, "push")
        self.assertEqual(push_spec.help, "Update remote refs")

        # Check status subcommand spec (with docstring)
        status_spec = cls.__subcommands__["status"]
        self.assertEqual(status_spec.name, "status")
        self.assertEqual(status_spec.help, "Show the working tree status.")
        self.assertIsNotNone(status_spec.description)
        self.assertIn("Displays paths", status_spec.description or "")


class GitArgumentsWithDirectClass(BaseArguments):
    """Test using argument_class parameter directly."""

    @subcommand(argument_class=GitCommitArguments, help="Commit changes")  # pyright: ignore[reportArgumentType]
    def commit(self):
        pass  # This function won't be called since argument_class is provided


class TestDirectArgumentClass(unittest.TestCase):
    def test_direct_argument_class(self):
        """Test using argument_class parameter directly."""
        args = GitArgumentsWithDirectClass(["commit", "-m", "direct test"])
        commit = args.last_subcommand
        assert isinstance(commit, GitCommitArguments), "commit should be a GitCommitArguments instance"
        self.assertEqual(commit.message.unwrap(), "direct test")


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Optional

from spargear import ArgumentSpec, BaseArguments, SubcommandSpec


class GitCommitArguments(BaseArguments):
    """Git commit command arguments."""

    message: ArgumentSpec[str] = ArgumentSpec(["-m", "--message"], required=True, help="Commit message")
    amend: ArgumentSpec[bool] = ArgumentSpec(["--amend"], action="store_true", help="Amend previous commit")


class GitPushArguments(BaseArguments):
    """Git push command arguments."""

    remote: ArgumentSpec[str] = ArgumentSpec(["remote"], nargs="?", default="origin", help="Remote name")
    branch: ArgumentSpec[Optional[str]] = ArgumentSpec(["branch"], nargs="?", help="Branch name")
    force: ArgumentSpec[bool] = ArgumentSpec(["-f", "--force"], action="store_true", help="Force push")


class GitArguments(BaseArguments):
    """Git command line interface example."""

    verbose: ArgumentSpec[bool] = ArgumentSpec(["-v", "--verbose"], action="store_true", help="Increase verbosity")
    commit_cmd = SubcommandSpec(name="commit", help="Record changes", argument_class=GitCommitArguments)
    push_cmd = SubcommandSpec(name="push", help="Update remote", argument_class=GitPushArguments)


class TestGitArguments(unittest.TestCase):
    def test_commit_subcommand(self):
        # commit requires -m
        with self.assertRaises(SystemExit):
            GitArguments(["commit"])
        commit = GitArguments(["commit", "-m", "fix"]).last_subcommand
        assert isinstance(commit, GitCommitArguments), "commit should be an instance of GitCommitArguments"
        self.assertEqual(commit.message.unwrap(), "fix")
        self.assertFalse(commit.amend.unwrap())

    def test_commit_with_amend(self):
        commit = GitArguments(["commit", "-m", "msg", "--amend"]).last_subcommand
        assert isinstance(commit, GitCommitArguments), "commit should be an instance of GitCommitArguments"
        self.assertTrue(commit.amend.unwrap())

    def test_push_subcommand_defaults(self):
        push = GitArguments(["push"]).last_subcommand
        assert isinstance(push, GitPushArguments), "push should be an instance of GitPushArguments"
        self.assertEqual(push.remote.unwrap(), "origin")
        self.assertIsNone(push.branch.value)
        self.assertFalse(push.force.unwrap())

    def test_push_with_overrides(self):
        push = GitArguments(["push", "upstream", "dev", "--force"]).last_subcommand
        assert isinstance(push, GitPushArguments), "push should be an instance of GitPushArguments"
        self.assertEqual(push.remote.unwrap(), "upstream")
        self.assertEqual(push.branch.unwrap(), "dev")
        self.assertTrue(push.force.unwrap())


class BazArgs(BaseArguments):
    qux: ArgumentSpec[str] = ArgumentSpec(["--qux"], help="qux argument")


class BarArgs(BaseArguments):
    baz = SubcommandSpec("baz", help="do baz", argument_class=BazArgs)


class RootArgs(BaseArguments):
    foo: ArgumentSpec[str] = ArgumentSpec(["foo"], help="foo argument")
    bar = SubcommandSpec("bar", help="do bar", argument_class=BarArgs)


class TestNestedSubcommands(unittest.TestCase):
    def test_two_levels(self):
        baz = RootArgs(["FOO_VAL", "bar", "baz", "--qux", "QUX_VAL"]).last_subcommand
        assert isinstance(baz, BazArgs), f"baz should be an instance of BarArgs: {type(baz)}"
        self.assertEqual(baz.qux.unwrap(), "QUX_VAL")

    def test_error_on_missing(self):
        with self.assertRaises(SystemExit):
            RootArgs([])  # missing foo positional
        with self.assertRaises(SystemExit):
            RootArgs(["FOO_VAL", "VAL", "bar"])  # missing baz sub-subcommand


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Test script for default factory functionality in spargear."""

import datetime
import unittest
import uuid
from typing import Callable, List, Optional, Union

from spargear import ArgumentSpec, BaseArguments


class MyArguments(BaseArguments):
    """Test arguments with default factory support."""

    # Regular default value
    name: str = "default_name"
    """Name of the user"""

    # Default factory using lambda
    timestamp: Union[str, Callable[[], str]] = (
        lambda: datetime.datetime.now().isoformat()
    )
    """Current timestamp (generated at parse time)"""

    # Default factory using function
    session_id: Union[str, Callable[[], str]] = lambda: str(uuid.uuid4())
    """Unique session ID (generated at parse time)"""

    # Explicit ArgumentSpec with default_factory
    log_file: ArgumentSpec[str] = ArgumentSpec(
        name_or_flags=["--log-file"],
        default_factory=lambda: f"log_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.txt",
        help="Log file name (auto-generated with timestamp)",
    )

    # Optional argument with no default
    config_file: Optional[str] = None
    """Optional configuration file"""

    # List with default factory
    tags: Union[List[str], Callable[[], List[str]]] = lambda: ["default", "auto"]
    """List of tags"""


class TestDefaultFactory(unittest.TestCase):
    def test_default_factory(self) -> None:
        """Test the default factory functionality."""
        print("Testing default factory functionality...")

        # Test 1: No arguments provided - should use default factories
        print("\n=== Test 1: No arguments ===")
        args1 = MyArguments([])
        print(f"name: {args1.get('name')}")
        print(f"timestamp: {args1.get('timestamp')}")
        print(f"session_id: {args1.get('session_id')}")
        print(f"log_file: {args1.get('log_file')}")
        print(f"config_file: {args1.get('config_file')}")
        print(f"tags: {args1.get('tags')}")

        # Test 2: Some arguments provided
        print("\n=== Test 2: Some arguments provided ===")
        args2 = MyArguments(["--name", "custom_name", "--config-file", "config.yaml"])
        print(f"name: {args2.get('name')}")
        print(f"timestamp: {args2.get('timestamp')}")
        print(f"session_id: {args2.get('session_id')}")
        print(f"log_file: {args2.get('log_file')}")
        print(f"config_file: {args2.get('config_file')}")
        print(f"tags: {args2.get('tags')}")

        # Test 3: Verify that default factories generate different values
        print("\n=== Test 3: Different instances have different generated values ===")
        args3 = MyArguments([])
        args4 = MyArguments([])

        print(f"args3 session_id: {args3.get('session_id')}")
        print(f"args4 session_id: {args4.get('session_id')}")
        print(
            f"Session IDs are different: {args3.get('session_id') != args4.get('session_id')}"
        )


if __name__ == "__main__":
    unittest.main()



import argparse
import ast
import inspect
import logging
import sys
import textwrap
import types
from functools import partial

if sys.version_info < (3, 9):
    import typing_extensions as typing

else:
    import typing

Annotated = typing.Annotated
get_type_hints = typing.get_type_hints

SUPPRESS_LITERAL_TYPE = typing.Literal["==SUPPRESS=="]
SUPPRESS: SUPPRESS_LITERAL_TYPE = "==SUPPRESS=="
ACTION_TYPES_THAT_DONT_SUPPORT_TYPE_KWARG = (
    "store_const",
    "store_true",
    "store_false",
    "append_const",
    "count",
    "help",
    "version",
)
Action = typing.Optional[
    typing.Literal[
        "store",
        "store_const",
        "store_true",
        "store_false",
        "append",
        "append_const",
        "count",
        "help",
        "version",
        "extend",
    ]
]
ContainerTypes = typing.Tuple[
    typing.Union[
        typing.Type[typing.List[object]], typing.Type[typing.Tuple[object, ...]]
    ],
    ...,
]

logger = logging.getLogger(__name__)


class FileProtocol(typing.Protocol):
    """A protocol that defines the methods expected from file-like objects."""

    def read(self, n: int = -1) -> str: ...
    def write(self, s: str) -> int: ...
    def close(self) -> None: ...


class TypedFileType:
    """A wrapper around argparse. FileType that returns FileProtocol compatible objects."""

    def __init__(
        self,
        mode: str,
        bufsize: int = -1,
        encoding: typing.Optional[str] = None,
        errors: typing.Optional[str] = None,
    ) -> None:
        self.file_type = argparse.FileType(mode, bufsize, encoding, errors)

    def __call__(self, string: str) -> typing.Union[typing.IO[str], typing.IO[bytes]]:
        return self.file_type(string)


def get_origin(obj: object) -> typing.Optional[object]:
    """Get the origin of a type, similar to typing.get_origin.

    e.g. typing.List[int] -> list
         typing.List -> None"""
    return typing.get_origin(obj)


def get_args(obj: object) -> typing.Tuple[object, ...]:
    """Get the arguments of a type, similar to typing.get_args."""
    return typing.get_args(obj)


def get_union_args(t: object) -> typing.Tuple[object, ...]:
    origin = get_origin(t)
    if origin is typing.Union:
        return get_args(t)
    if sys.version_info >= (3, 10) and origin is types.UnionType:
        return get_args(t)
    return ()


def is_optional(t: object) -> bool:
    """Check if a type is Optional."""
    return type(None) in get_union_args(t)


def sanitize_name(name: str) -> str:
    """Sanitize a name for use as a command-line argument."""
    return name.replace("_", "-").lower().lstrip("-")


def sanitize_flag(name: str) -> str:
    """Sanitize a name for use as a command-line argument."""
    if name.isupper():
        # if the name is all uppercase, assume it's positional
        return sanitize_name(name)
    else:
        return f"--{sanitize_name(name)}"  # if the name is not all uppercase, assume it's a flag


def ensure_no_optional(t: object) -> object:
    """Ensure that the type is not Optional."""
    non_none_args = tuple(arg for arg in get_union_args(t) if arg is not type(None))
    if not non_none_args:
        return t
    if len(non_none_args) == 1:
        return non_none_args[0]
    return typing.Union[non_none_args]  # pyright: ignore[reportInvalidTypeArguments]


def get_arguments_of_container_types(
    type_no_optional_or_spec: object, container_types: ContainerTypes
) -> typing.Optional[typing.Tuple[object, ...]]:
    if isinstance(type_no_optional_or_spec, type) and issubclass(
        type_no_optional_or_spec, container_types
    ):
        return ()

    type_no_optional_or_spec = typing.cast(object, type_no_optional_or_spec)
    type_no_optional_or_spec_origin: typing.Optional[object] = get_origin(
        type_no_optional_or_spec
    )
    if isinstance(type_no_optional_or_spec_origin, type) and issubclass(
        type_no_optional_or_spec_origin, container_types
    ):
        return get_args(type_no_optional_or_spec)
    return None


def get_type_of_element_of_container_types(
    type_no_optional_or_spec: object, container_types: ContainerTypes
) -> typing.Optional[type]:
    iterable_arguments = get_arguments_of_container_types(
        type_no_optional_or_spec=type_no_optional_or_spec,
        container_types=container_types,
    )
    if iterable_arguments is None:
        return None
    else:
        return next((it for it in iterable_arguments if isinstance(it, type)), None)


def get_literals(
    type_no_optional_or_spec: object, container_types: ContainerTypes
) -> typing.Optional[typing.Tuple[object, ...]]:
    """Get the literals of the list element type."""
    if get_origin(type_no_optional_or_spec) is typing.Literal:
        # Extract literals from Literal type
        return get_args(type_no_optional_or_spec)

    elif (
        arguments_of_container_types := get_arguments_of_container_types(
            type_no_optional_or_spec=type_no_optional_or_spec,
            container_types=container_types,
        )
    ) and get_origin(
        first_argument_of_container_types := arguments_of_container_types[0]
    ) is typing.Literal:
        # Extract literals from List[Literal] or Tuple[Literal, ...]
        return get_args(first_argument_of_container_types)
    return None


def extract_attr_docstrings(cls: typing.Type[object]) -> typing.Dict[str, str]:
    """
    Extracts docstrings from class attributes.
    This function inspects the class definition and retrieves the docstrings
    associated with each attribute.
    """
    try:
        source = inspect.getsource(cls)
        source_ast = ast.parse(textwrap.dedent(source))

        docstrings: typing.Dict[str, str] = {}
        last_attr: typing.Optional[str] = None

        class_def = next(
            (node for node in source_ast.body if isinstance(node, ast.ClassDef)), None
        )
        if class_def is None:
            return {}

        for node in class_def.body:
            # Annotated assignment (e.g., `a: int`)
            if isinstance(node, ast.AnnAssign) and isinstance(node.target, ast.Name):
                last_attr = node.target.id

            # """docstring"""
            elif isinstance(node, ast.Expr) and isinstance(node.value, ast.Constant):
                if isinstance(node.value.value, str) and last_attr:
                    docstrings[last_attr] = node.value.value.strip()
                    last_attr = None
            else:
                last_attr = None  # cut off if we see something else

        return docstrings
    except Exception as e:
        logger.debug(f"Failed to extract docstrings from {cls.__name__}: {e}")
        return {}


def unwrap_callable(func: typing.Callable[..., object]) -> typing.Callable[..., object]:
    """
    Get the name of a callable.
    This function is used to get the name of a callable, even if it is wrapped by a decorator.
    It is used to get the name of a subcommand.
    """
    # 1) staticmethod / classmethod → unwrap __func__
    if isinstance(func, (staticmethod, classmethod)):
        func = func.__func__  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]

    # 2) bound method(A().method) → unwrap __func__
    func = getattr(func, "__func__", func)

    # 3) functools.partial → unwrap func
    if isinstance(func, partial):
        func = func.func

    # 4) unwrap decorator chain (functools.wraps based)
    func = inspect.unwrap(func)
    return func


def assert_type(val: object, type_: type) -> None:
    """Assert that the value is of the given type."""
    
    if not isinstance(val, type_):
        raise TypeError(f"Value `{val}` is not of type `{type_}`")



import argparse
from dataclasses import dataclass, field, fields
from typing import (
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from ._typing import (
    ACTION_TYPES_THAT_DONT_SUPPORT_TYPE_KWARG,
    SUPPRESS_LITERAL_TYPE,
    Action,
    TypedFileType,
    ensure_no_optional,
    get_args,
    get_arguments_of_container_types,
    get_literals,
    get_origin,
    get_type_of_element_of_container_types,
)

T = TypeVar("T")


@dataclass
class ArgumentSpec(Generic[T]):
    """Represents the specification for a command-line argument."""

    name_or_flags: List[str]
    action: Action = None
    nargs: Optional[Union[int, Literal["*", "+", "?"]]] = None
    const: Optional[T] = None
    default: Optional[Union[T, SUPPRESS_LITERAL_TYPE]] = None
    default_factory: Optional[Callable[[], T]] = None
    choices: Optional[Sequence[T]] = None
    required: bool = False
    help: str = ""
    metavar: Optional[str] = None
    version: Optional[str] = None
    type: Optional[
        Union[Callable[[str], T], Type[argparse.FileType], TypedFileType]
    ] = None
    dest: Optional[str] = None
    value: Optional[T] = field(init=False, default=None)  # Parsed value stored here

    def __post_init__(self) -> None:
        """Validate that default and default_factory are not both set."""
        if self.default is not None and self.default_factory is not None:
            raise ValueError("Cannot specify both 'default' and 'default_factory'")

    def unwrap(self) -> T:
        """Returns the value, raising an error if it's None."""
        if self.value is None:
            raise ValueError(f"Value for {self.name_or_flags} is None.")
        return self.value

    def get_add_argument_kwargs(self) -> Dict[str, object]:
        """Prepares keyword arguments for argparse.ArgumentParser.add_argument."""
        kwargs: Dict[str, object] = {}
        argparse_fields: Set[str] = {
            f.name
            for f in fields(self)
            if f.name not in ("name_or_flags", "value", "default_factory")
        }
        for field_name in argparse_fields:
            attr_value: object = getattr(self, field_name)
            if field_name == "default":
                if attr_value is None:
                    # If we have a default_factory, don't set default in argparse
                    if self.default_factory is not None:
                        kwargs[field_name] = argparse.SUPPRESS
                    else:
                        pass  # Keep default=None if explicitly set or inferred
                elif attr_value in get_args(SUPPRESS_LITERAL_TYPE):
                    kwargs[field_name] = argparse.SUPPRESS
                else:
                    kwargs[field_name] = attr_value
            elif attr_value is not None:
                if (
                    field_name == "type"
                    and self.action in ACTION_TYPES_THAT_DONT_SUPPORT_TYPE_KWARG
                ):
                    continue
                kwargs[field_name] = attr_value
        return kwargs

    def apply_default_factory(self) -> None:
        """Apply the default factory if value is None and default_factory is set."""
        if self.value is None and self.default_factory is not None:
            self.value = self.default_factory()


class ArgumentSpecType(NamedTuple):
    """Represents the type information extracted from ArgumentSpec type hints."""

    type_no_optional_or_spec: object  # The T in ArgumentSpec[T]
    is_specless_type: bool = False

    @classmethod
    def from_type_hint(cls, type_hint: object):
        """Extract type information from a type hint."""
        type_no_spec: object = ensure_no_argspec(type_hint)  # pyright: ignore[reportPrivateUsage]
        return cls(
            type_no_optional_or_spec=ensure_no_optional(type_no_spec),
            is_specless_type=type_hint is type_no_spec,
        )

    @property
    def choices(self) -> Optional[Tuple[object, ...]]:
        """Extract choices from Literal types."""
        return get_literals(
            type_no_optional_or_spec=self.type_no_optional_or_spec,
            container_types=(list, tuple),
        )

    @property
    def type(self) -> Optional[Type[object]]:
        """Determine the appropriate type for the argument."""
        t = get_type_of_element_of_container_types(
            type_no_optional_or_spec=self.type_no_optional_or_spec,
            container_types=(list, tuple),
        )
        if t is not None:
            return t
        if isinstance(self.type_no_optional_or_spec, type):
            return self.type_no_optional_or_spec
        return None

    @property
    def should_return_as_list(self) -> bool:
        """Determines if the argument should be returned as a list."""
        return (
            get_arguments_of_container_types(
                type_no_optional_or_spec=self.type_no_optional_or_spec,
                container_types=(list,),
            )
            is not None
        )

    @property
    def should_return_as_tuple(self) -> bool:
        """Determines if the argument should be returned as a tuple."""
        return (
            get_arguments_of_container_types(
                type_no_optional_or_spec=self.type_no_optional_or_spec,
                container_types=(tuple,),
            )
            is not None
        )

    @property
    def tuple_nargs(self) -> Optional[Union[int, Literal["+"]]]:
        """Determine the number of arguments for a tuple type."""
        if self.should_return_as_tuple and (
            args := get_args(self.type_no_optional_or_spec)
        ):
            if Ellipsis not in args:
                return len(args)
            else:
                return "+"
        return None

    @property
    def basic_info(self) -> Dict[str, object]:
        """Returns a dictionary with basic information about the argument."""
        return {
            "type_no_optional_or_spec": self.type_no_optional_or_spec,
            "is_specless_type": self.is_specless_type,
            "choices": self.choices,
            "type": self.type,
            "should_return_as_list": self.should_return_as_list,
            "should_return_as_tuple": self.should_return_as_tuple,
            "tuple_nargs": self.tuple_nargs,
        }


def ensure_no_argspec(t: object) -> object:
    """Unwraps the ArgumentSpec type to get the actual type."""
    if (
        (origin := get_origin(t)) is not None
        and isinstance(origin, type)
        and issubclass(origin, ArgumentSpec)
        and (args := get_args(t))
    ):
        # Extract T from ArgumentSpec[T]
        return args[0]
    return t



from dataclasses import dataclass, field
from inspect import getdoc
from typing import TYPE_CHECKING, Callable, Generic, Optional, Type, TypeVar

from ._typing import sanitize_name, unwrap_callable

if TYPE_CHECKING:
    from .base import BaseArguments

S = TypeVar("S", bound="BaseArguments")


@dataclass
class SubcommandSpec(Generic[S]):
    """Represents a subcommand specification for command-line interfaces."""

    name: str
    """The name of the subcommand."""
    argument_class: Optional[Type[S]] = None
    """The BaseArguments subclass that defines the subcommand's arguments."""
    argument_class_factory: Optional[Callable[[], Type[S]]] = None
    """A factory function that returns the BaseArguments subclass."""
    help: str = ""
    """Brief help text for the subcommand."""
    description: Optional[str] = None
    """Detailed description of the subcommand."""

    # Private field to cache the result of factory function
    _cached_argument_class: Optional[Type[S]] = field(
        default=None, init=False, repr=False
    )

    def __post_init__(self) -> None:
        """Validate that either argument_class or argument_class_factory is provided."""
        if self.argument_class is None and self.argument_class_factory is None:
            raise ValueError(
                "Either argument_class or argument_class_factory must be provided"
            )
        if self.argument_class is not None and self.argument_class_factory is not None:
            raise ValueError(
                "Only one of argument_class or argument_class_factory should be provided"
            )

    def get_argument_class(self) -> Type[S]:
        """Get the argument class, either directly or from the factory."""
        if self.argument_class is not None:
            return self.argument_class
        elif self.argument_class_factory is not None:
            # Use cached result if available
            if self._cached_argument_class is not None:
                return self._cached_argument_class
            # Call factory and cache the result
            self._cached_argument_class = self.argument_class_factory()
            return self._cached_argument_class
        else:
            raise ValueError("No argument class or factory available")


def subcommand(
    name: Optional[str] = None,
    help: str = "",
    description: Optional[str] = None,
    argument_class: Optional[Type[S]] = None,
) -> Callable[[Callable[..., Type[S]]], SubcommandSpec[S]]:
    """
    Decorator to automatically create SubcommandSpec from a method.

    This is the recommended way to define subcommands in spargear. The decorator
    automatically handles method-to-factory conversion and extracts metadata
    from the method name and docstring.

    Args:
        name: The name of the subcommand. If not provided, uses the method name.
        help: Brief help text for the subcommand. If not provided, uses the first
              line of the method's docstring.
        description: Detailed description of the subcommand. If not provided,
                    uses the remaining lines of the method's docstring.
        argument_class: The BaseArguments subclass. If provided, the method won't
                       be called and this class will be used directly.

    Returns:
        A SubcommandSpec instance that can be used as a class attribute.

    Examples:
        Basic usage:

        ```python
        class MyApp(BaseArguments):
            @subcommand(help="Initialize a new project")
            def init():
                return InitArguments
        ```

        With custom name and docstring help:

        ```python
        class MyApp(BaseArguments):
            @subcommand(name="db-migrate")
            def database_migrate():
                '''Run database migrations.

                This command applies all pending database migrations
                to bring the database schema up to date.
                '''
                return MigrateArguments
        ```

        With direct argument class (method won't be called):

        ```python
        class MyApp(BaseArguments):
            @subcommand(argument_class=ServeArguments, help="Start server")
            def serve():
                pass  # This method body is ignored
        ```

    Note:
        - No need for @staticmethod - the decorator handles method calling automatically
        - The method should return a BaseArguments subclass
        - Method docstrings are automatically used for help and description
        - Method names are automatically converted to subcommand names
    """

    def decorator(func: Callable[..., Type[S]]) -> SubcommandSpec[S]:
        # Handle staticmethod objects
        unwrapped_func = unwrap_callable(func)

        # Extract name from function name if not provided
        subcommand_name: str = name or sanitize_name(unwrapped_func.__name__)

        # Extract help from docstring if not provided
        func_help: str = help
        func_description: Optional[str] = description
        if not func_help and (doc := getdoc(unwrapped_func)):
            # Use first line of docstring as help
            lines = doc.strip().split("\n")
            func_help = lines[0].strip()
            # Use rest of docstring as description if available and description not provided
            if not func_description and len(lines) > 1:
                func_description = "\n".join(line.strip() for line in lines[1:]).strip()
                if not func_description:
                    func_description = None

        # Determine argument_class or argument_class_factory
        if argument_class is not None:
            # Use provided argument_class directly
            return SubcommandSpec(
                name=subcommand_name,
                argument_class=argument_class,
                help=func_help,
                description=func_description,
            )
        else:
            # Always use the function as a factory
            def argument_class_factory() -> Type[S]:
                # At class definition time, methods are unbound functions
                # So we can call them directly without self
                try:
                    result = unwrapped_func()
                    return result  # pyright: ignore[reportReturnType]
                except TypeError:
                    # If there's still a TypeError, the function probably has other required parameters
                    raise ValueError(
                        f"Method '{unwrapped_func.__name__}' cannot be used as a factory. Make sure it has no required parameters (including 'self')."
                    )

            return SubcommandSpec(
                name=subcommand_name,
                argument_class_factory=argument_class_factory,
                help=func_help,
                description=func_description,
            )

    return decorator



import unittest
from typing import List, Type

from spargear import ArgumentSpec, BaseArguments, SubcommandSpec


class DynamicCommitArguments(BaseArguments):
    """Dynamically created commit command arguments."""

    message: ArgumentSpec[str] = ArgumentSpec(
        ["-m", "--message"], required=True, help="Commit message"
    )
    amend: ArgumentSpec[bool] = ArgumentSpec(
        ["--amend"], action="store_true", help="Amend previous commit"
    )


class DynamicPushArguments(BaseArguments):
    """Dynamically created push command arguments."""

    remote: ArgumentSpec[str] = ArgumentSpec(
        ["remote"], nargs="?", default="origin", help="Remote name"
    )
    force: ArgumentSpec[bool] = ArgumentSpec(
        ["-f", "--force"], action="store_true", help="Force push"
    )


def create_commit_class() -> Type[DynamicCommitArguments]:
    """Factory function to create commit arguments class."""
    return DynamicCommitArguments


def create_push_class() -> Type[BaseArguments]:
    """Factory function to create push arguments class."""
    return DynamicPushArguments


class DynamicGitArguments(BaseArguments):
    """Git CLI with dynamic subcommand creation."""

    verbose: ArgumentSpec[bool] = ArgumentSpec(
        ["-v", "--verbose"], action="store_true", help="Increase verbosity"
    )

    # Using factory function
    commit_cmd: SubcommandSpec[DynamicCommitArguments] = SubcommandSpec(
        name="commit",
        argument_class_factory=create_commit_class,
        help="Record changes",
    )

    # Using lambda factory
    push_cmd: SubcommandSpec[DynamicPushArguments] = SubcommandSpec(
        name="push",
        argument_class_factory=lambda: DynamicPushArguments,
        help="Update remote",
    )

    # Using direct class (should still work)
    status_cmd: SubcommandSpec[DynamicCommitArguments] = SubcommandSpec(
        name="status",
        argument_class=DynamicCommitArguments,  # Reusing for simplicity
        help="Show status",
    )


class TestDynamicSubcommands(unittest.TestCase):
    def test_factory_function_commit(self):
        """Test subcommand with factory function."""
        # commit requires -m
        with self.assertRaises(SystemExit):
            DynamicGitArguments(["commit"])

        git_args = DynamicGitArguments(["commit", "-m", "fix"])
        commit = git_args.last_subcommand
        assert isinstance(commit, DynamicCommitArguments), (
            "commit should be an instance of DynamicCommitArguments"
        )
        self.assertEqual(commit.message.unwrap(), "fix")
        self.assertFalse(commit.amend.unwrap())

    def test_lambda_factory_push(self):
        """Test subcommand with lambda factory."""
        git_args = DynamicGitArguments(["push"])
        push = git_args.last_subcommand
        assert isinstance(push, DynamicPushArguments), (
            "push should be an instance of DynamicPushArguments"
        )
        self.assertEqual(push.remote.unwrap(), "origin")
        self.assertFalse(push.force.unwrap())

    def test_direct_class_status(self):
        """Test subcommand with direct class (should still work)."""
        git_args = DynamicGitArguments(["status", "-m", "test"])
        status = git_args.last_subcommand
        assert isinstance(status, DynamicCommitArguments), (
            "status should be an instance of DynamicCommitArguments"
        )
        self.assertEqual(status.message.unwrap(), "test")

    def test_factory_called_on_demand(self):
        """Test that factory is called on demand, not at class definition time."""
        calls: List[str] = []

        def counting_factory() -> Type[DynamicCommitArguments]:
            calls.append("called")
            return DynamicCommitArguments

        class TestArgs(BaseArguments):
            test_cmd: SubcommandSpec[DynamicCommitArguments] = SubcommandSpec(
                name="test",
                argument_class_factory=counting_factory,
                help="Test command",
            )

        # Factory should not be called yet
        self.assertEqual(len(calls), 0)

        # Factory should be called when getting argument class
        TestArgs(["test", "-m", "message"])
        self.assertEqual(len(calls), 1)


class TestSubcommandSpecValidation(unittest.TestCase):
    def test_no_class_or_factory_raises_error(self):
        """Test that providing neither argument_class nor argument_class_factory raises an error."""
        with self.assertRaises(ValueError) as cm:
            SubcommandSpec(name="test")
        self.assertIn(
            "Either argument_class or argument_class_factory must be provided",
            str(cm.exception),
        )

    def test_both_class_and_factory_raises_error(self):
        """Test that providing both argument_class and argument_class_factory raises an error."""
        with self.assertRaises(ValueError) as cm:
            SubcommandSpec(
                name="test",
                argument_class=DynamicCommitArguments,
                argument_class_factory=create_commit_class,
            )
        self.assertIn(
            "Only one of argument_class or argument_class_factory should be provided",
            str(cm.exception),
        )

    def test_get_argument_class_with_class(self):
        """Test get_argument_class with direct class."""
        spec = SubcommandSpec(name="test", argument_class=DynamicCommitArguments)
        self.assertEqual(spec.get_argument_class(), DynamicCommitArguments)

    def test_get_argument_class_with_factory(self):
        """Test get_argument_class with factory function."""
        spec = SubcommandSpec(name="test", argument_class_factory=create_commit_class)
        self.assertEqual(spec.get_argument_class(), DynamicCommitArguments)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Test script for serialization and dataclass conversion features."""

import os
import tempfile
import unittest
import uuid
from typing import Dict, List, Optional

from spargear import ArgumentSpec, BaseArguments


class ConfigExample(BaseArguments):
    """Example configuration with various argument types."""

    # Basic arguments
    name: str = "default_app"
    """Application name"""

    port: int = 8080
    """Server port"""

    debug: bool = False
    """Enable debug mode"""

    # Optional arguments
    database_url: Optional[str] = None
    """Database connection URL"""

    # List arguments
    allowed_hosts: List[str] = ["localhost", "127.0.0.1"]
    """List of allowed hosts"""

    # ArgumentSpec with default factory
    session_id: ArgumentSpec[str] = ArgumentSpec(
        name_or_flags=["--session-id"],
        default_factory=lambda: str(uuid.uuid4()),
        help="Unique session identifier",
    )

    # ArgumentSpec with regular default
    log_level: ArgumentSpec[str] = ArgumentSpec(
        name_or_flags=["--log-level"],
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        default="INFO",
        help="Logging level",
    )


class TestSerialization(unittest.TestCase):
    def test_dataclass_conversion(self) -> None:
        """Test converting BaseArguments to dataclass."""
        print("=== Testing Dataclass Conversion ===")

        # Create instance with some arguments
        config = ConfigExample(["--name", "myapp", "--port", "9000", "--debug"])

        # Convert to dataclass
        config_dc = config.to_dataclass()
        print(f"Dataclass type: {type(config_dc)}")
        print(f"Dataclass name: {config_dc.name}")
        print(f"Dataclass port: {config_dc.port}")
        print(f"Dataclass debug: {config_dc.debug}")
        print(f"Dataclass session_id: {config_dc.session_id}")
        print()

    def test_dict_conversion(self) -> None:
        """Test converting BaseArguments to dictionary."""
        print("=== Testing Dictionary Conversion ===")

        config = ConfigExample(["--name", "testapp", "--log-level", "DEBUG"])
        config_dict = config.to_dict()

        print("Configuration as dictionary:")
        for key, value in config_dict.items():
            print(f"  {key}: {value}")
        print()

    def test_json_serialization(self) -> None:
        """Test JSON serialization and deserialization."""
        print("=== Testing JSON Serialization ===")

        # Create original config
        config1 = ConfigExample(["--name", "jsonapp", "--port", "3000"])

        # Serialize to JSON string
        json_str = config1.to_json()
        print("JSON representation:")
        print(json_str)
        print()

        # Save to file
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            json_str = config1.to_json()
            f.write(json_str)
            json_file = f.name

        try:
            # Load from file
            config2 = ConfigExample.load_config(json_file)
            print("Loaded from JSON file:")
            print(f"  name: {config2.name}")
            print(f"  port: {config2.port}")
            print(f"  session_id: {config2.session_id.unwrap()}")
            print()

            # Test from_json with string
            config3 = ConfigExample.from_json(json_str)
            print("Loaded from JSON string:")
            print(f"  name: {config3.name}")
            print(f"  debug: {config3.debug}")
            print()

        finally:
            os.unlink(json_file)

    def test_pickle_serialization(self) -> None:
        """Test pickle serialization and deserialization."""
        print("=== Testing Pickle Serialization ===")

        config1 = ConfigExample(["--name", "pickleapp", "--debug"])
        print(config1)  ###

        # Save to pickle file
        with tempfile.NamedTemporaryFile(suffix=".pkl", delete=False) as f:
            f.write(config1.to_pickle())
            pickle_file = f.name

        try:
            # Load from pickle file
            config2 = ConfigExample.load_config(pickle_file)
            print("Loaded from pickle file:")
            print(f"  name: {config2.name}")
            print(f"  debug: {config2.debug}")
            print(f"  session_id: {config2.session_id.unwrap()}")
            print()

        finally:
            os.unlink(pickle_file)

    def test_update_functionality(self) -> None:
        """Test updating configuration from dictionary."""
        print("=== Testing Update Functionality ===")

        config = ConfigExample([])
        print("Original config:")
        print(f"  name: {config.name}")
        print(f"  port: {config.port}")

        # Update from dictionary
        updates: Dict[str, object] = {
            "name": "updated_app",
            "port": 5000,
            "debug": True,
        }
        config.update_from_dict(updates)

        print("After update:")
        print(f"  name: {config.name}")
        print(f"  port: {config.port}")
        print(f"  debug: {config.debug}")
        print()

    def test_command_line_override(self) -> None:
        """Test loading config with command line override."""
        print("=== Testing Command Line Override ===")

        # Create config data
        config_data: Dict[str, object] = {
            "name": "config_app",
            "port": 4000,
            "debug": False,
        }

        # Create instance with command line args that override config
        config = ConfigExample.from_dict(config_data, args=["--port", "6000", "--debug"])

        print("Config with command line override:")
        print(f"  name: {config.name} (from config)")
        print(f"  port: {config.port} (overridden by command line)")
        print(f"  debug: {config.debug} (overridden by command line)")
        print()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Simple test for default factory functionality."""

import unittest
import uuid
from typing import Union, Callable
from spargear import BaseArguments


class SimpleArgs(BaseArguments):
    """Simple test arguments."""

    session_id: Union[str, Callable[[], str]] = lambda: str(uuid.uuid4())
    """Unique session ID"""


class TestSimpleCallable(unittest.TestCase):
    def test_simple_callable(self) -> None:
        print("Creating first instance...")
        args1 = SimpleArgs([])
        print(f"args1 session_id: {args1.get('session_id')}")

        print("Creating second instance...")
        args2 = SimpleArgs([])
        print(f"args2 session_id: {args2.get('session_id')}")

        print(
            f"Are they different? {args1.get('session_id') != args2.get('session_id')}"
        )


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import cast

from spargear import ArgumentSpec, BaseArguments, SubcommandSpec


class SharedSubCommand(BaseArguments):
    """공유되는 서브커맨드 클래스 - 여러 부모에서 사용됨"""

    shared_arg: ArgumentSpec[str] = ArgumentSpec(["--shared"], default="default", help="Shared argument")
    value: ArgumentSpec[int] = ArgumentSpec(["--value"], type=int, default=0, help="Value argument")


class ParentA(BaseArguments):
    """첫 번째 부모 클래스"""

    parent_a_arg: ArgumentSpec[str] = ArgumentSpec(["--parent-a"], default="a", help="Parent A argument")
    shared_sub = SubcommandSpec("shared", help="Shared subcommand", argument_class=SharedSubCommand)


class ParentB(BaseArguments):
    """두 번째 부모 클래스"""

    parent_b_arg: ArgumentSpec[str] = ArgumentSpec(["--parent-b"], default="b", help="Parent B argument")
    shared_sub = SubcommandSpec("shared", help="Shared subcommand", argument_class=SharedSubCommand)


class TestInstanceIsolation(unittest.TestCase):
    """인스턴스 격리 테스트"""

    def test_shared_subcommand_class_isolation(self):
        """동일한 서브커맨드 클래스가 여러 부모에서 사용될 때 격리되는지 테스트"""

        # 첫 번째 부모에서 서브커맨드 사용
        parent_a = ParentA(["shared", "--shared", "value_a", "--value", "100"])
        sub_a = parent_a.last_subcommand

        # 두 번째 부모에서 서브커맨드 사용
        parent_b = ParentB(["shared", "--shared", "value_b", "--value", "200"])
        sub_b = parent_b.last_subcommand

        # 타입 확인
        self.assertIsInstance(sub_a, SharedSubCommand)
        self.assertIsInstance(sub_b, SharedSubCommand)
        self.assertIsNotNone(sub_a)
        self.assertIsNotNone(sub_b)

        # 타입 캐스팅으로 타입 체커 오류 해결
        sub_a_typed = cast(SharedSubCommand, sub_a)
        sub_b_typed = cast(SharedSubCommand, sub_b)

        # 값 격리 확인
        self.assertEqual(sub_a_typed.shared_arg.unwrap(), "value_a")
        self.assertEqual(sub_a_typed.value.unwrap(), 100)

        self.assertEqual(sub_b_typed.shared_arg.unwrap(), "value_b")
        self.assertEqual(sub_b_typed.value.unwrap(), 200)

        # 인스턴스가 독립적인지 확인
        self.assertNotEqual(sub_a_typed.shared_arg.unwrap(), sub_b_typed.shared_arg.unwrap())
        self.assertNotEqual(sub_a_typed.value.unwrap(), sub_b_typed.value.unwrap())

    def test_multiple_instances_same_class(self):
        """같은 클래스의 여러 인스턴스가 서로 독립적인지 테스트"""

        # 같은 클래스로 여러 인스턴스 생성
        instance1 = ParentA(["shared", "--shared", "inst1", "--value", "10"])
        instance2 = ParentA(["shared", "--shared", "inst2", "--value", "20"])
        instance3 = ParentA(["shared", "--shared", "inst3", "--value", "30"])

        sub1 = instance1.last_subcommand
        sub2 = instance2.last_subcommand
        sub3 = instance3.last_subcommand

        # 타입 확인
        self.assertIsInstance(sub1, SharedSubCommand)
        self.assertIsInstance(sub2, SharedSubCommand)
        self.assertIsInstance(sub3, SharedSubCommand)

        # 타입 캐스팅
        sub1_typed = cast(SharedSubCommand, sub1)
        sub2_typed = cast(SharedSubCommand, sub2)
        sub3_typed = cast(SharedSubCommand, sub3)

        # 각 인스턴스가 독립적인 값을 가지는지 확인
        self.assertEqual(sub1_typed.shared_arg.unwrap(), "inst1")
        self.assertEqual(sub1_typed.value.unwrap(), 10)

        self.assertEqual(sub2_typed.shared_arg.unwrap(), "inst2")
        self.assertEqual(sub2_typed.value.unwrap(), 20)

        self.assertEqual(sub3_typed.shared_arg.unwrap(), "inst3")
        self.assertEqual(sub3_typed.value.unwrap(), 30)

        # 서로 다른 인스턴스인지 확인
        self.assertIsNot(sub1, sub2)
        self.assertIsNot(sub2, sub3)
        self.assertIsNot(sub1, sub3)

    def test_argumentspec_instance_isolation(self):
        """ArgumentSpec 인스턴스가 격리되는지 테스트"""

        parent1 = ParentA(["--parent-a", "custom1"])
        parent2 = ParentA(["--parent-a", "custom2"])

        # ArgumentSpec 인스턴스가 독립적인지 확인
        self.assertEqual(parent1.parent_a_arg.unwrap(), "custom1")
        self.assertEqual(parent2.parent_a_arg.unwrap(), "custom2")

        # spec 객체 자체는 다른 인스턴스여야 함
        self.assertIsNot(parent1.parent_a_arg, parent2.parent_a_arg)

        # 값 변경이 다른 인스턴스에 영향을 주지 않는지 확인
        parent1.parent_a_arg.value = "modified1"
        parent2.parent_a_arg.value = "modified2"

        self.assertEqual(parent1.parent_a_arg.value, "modified1")
        self.assertEqual(parent2.parent_a_arg.value, "modified2")

    def test_default_values_isolation(self):
        """기본값이 독립적으로 관리되는지 테스트"""

        # 기본값 사용
        parent1 = ParentA([])
        parent2 = ParentA([])

        # 기본값이 같은지 확인
        self.assertEqual(parent1.parent_a_arg.unwrap(), "a")
        self.assertEqual(parent2.parent_a_arg.unwrap(), "a")

        # 하나의 값을 변경
        parent1.parent_a_arg.value = "changed"

        # 다른 인스턴스에 영향을 주지 않는지 확인
        self.assertEqual(parent1.parent_a_arg.value, "changed")
        self.assertEqual(parent2.parent_a_arg.unwrap(), "a")

    def test_class_attribute_preservation(self):
        """클래스 수준 속성이 보존되는지 테스트 - 회귀 테스트"""

        # 원래 클래스 속성 저장
        original_subcommands_a = ParentA.__subcommands__.copy()
        original_subcommands_b = ParentB.__subcommands__.copy()
        original_shared_subcommands = SharedSubCommand.__subcommands__.copy()

        # 여러 인스턴스 생성
        for i in range(5):
            ParentA(["shared", "--shared", f"test_{i}"])
            ParentB(["shared", "--shared", f"test_{i}"])

        # 클래스 속성이 변경되지 않았는지 확인
        # (이전에는 __parent__ 설정으로 인해 문제가 있었음)
        self.assertEqual(ParentA.__subcommands__, original_subcommands_a)
        self.assertEqual(ParentB.__subcommands__, original_subcommands_b)
        self.assertEqual(SharedSubCommand.__subcommands__, original_shared_subcommands)


if __name__ == "__main__":
    unittest.main()



import os
import tempfile
import unittest
from typing import List, Literal, Optional, Tuple

from spargear import SUPPRESS, FileProtocol, TypedFileType, ArgumentSpec, BaseArguments


class SimpleArguments(BaseArguments):
    """Example argument parser demonstrating various features."""

    my_str_arg: ArgumentSpec[str] = ArgumentSpec(["-s", "--string-arg"], default="Hello", help="A string argument.", metavar="TEXT")
    my_int_arg: ArgumentSpec[int] = ArgumentSpec(["-i", "--integer-arg"], help="A required integer argument.")
    verbose: ArgumentSpec[bool] = ArgumentSpec(["-v", "--verbose"], action="store_true", help="Increase output verbosity.")
    my_list_arg: ArgumentSpec[List[str]] = ArgumentSpec(["--list-values"], nargs=3, help="One or more values.", default=None)
    input_file: ArgumentSpec[FileProtocol] = ArgumentSpec(["input_file"], type=TypedFileType("r", encoding="utf-8"), help="Input file", metavar="INPUT")
    output_file: ArgumentSpec[Optional[FileProtocol]] = ArgumentSpec(["output_file"], type=TypedFileType("w", encoding="utf-8"), nargs="?", default=None, help="Output file")
    log_level: ArgumentSpec[Literal["DEBUG", "INFO", "WARNING", "ERROR"]] = ArgumentSpec(["--log-level"], default="INFO", help="Set log level.")
    mode: ArgumentSpec[Literal["fast", "slow", "careful"]] = ArgumentSpec(["--mode"], default="fast", help="Mode")
    enabled_features: ArgumentSpec[List[Literal["CACHE", "LOGGING", "RETRY"]]] = ArgumentSpec(["--features"], nargs="*", default=[], help="Enable features")
    tuple_features: ArgumentSpec[Tuple[Literal["CACHE", "LOGGING", "RETRY"], Literal["CACHE", "LOGGING", "RETRY"]]] = ArgumentSpec(["--tuple-features"], help="Tuple features")
    optional_flag: ArgumentSpec[str] = ArgumentSpec(["--opt-flag"], default=SUPPRESS, help="Optional flag suppressed if missing")


# raise Exception(SimpleArguments.__arguments__)


class TestSimpleArguments(unittest.TestCase):
    def test_missing_required(self):
        parser = SimpleArguments.get_parser()
        with self.assertRaises(SystemExit):
            parser.parse_args([])  # my_int_arg and input_file are required positional/required

    def test_basic_parsing_and_defaults(self):
        temp_in = tempfile.NamedTemporaryFile(mode="w+", delete=False, suffix=".txt")
        temp_out = tempfile.NamedTemporaryFile(mode="w+", delete=False, suffix=".txt")

        temp_in_path = temp_in.name
        temp_out_path = temp_out.name

        # 파일 닫기
        temp_in.close()
        temp_out.close()

        argv = [
            "-i",
            "42",
            "-s",
            "World",
            "--verbose",
            "--list-values",
            "a",
            "b",
            "c",
            temp_in_path,
            temp_out_path,
            "--log-level",
            "DEBUG",
            "--mode",
            "careful",
            "--features",
            "CACHE",
            "RETRY",
            "--tuple-features",
            "CACHE",
            "LOGGING",
        ]
        simple_args = SimpleArguments(argv)

        # 파일 객체가 열려 있다면 명시적으로 닫기
        input_file = simple_args.input_file.unwrap()
        output_file = simple_args.output_file.unwrap()
        input_file.close()
        os.remove(temp_in_path)  # 임시 파일 삭제
        if output_file is not None:
            output_file.close()
            os.remove(temp_out_path)

        self.assertEqual(simple_args.my_int_arg.unwrap(), 42)
        self.assertEqual(simple_args.my_str_arg.unwrap(), "World")
        self.assertTrue(simple_args.verbose.unwrap())

        self.assertListEqual(simple_args.my_list_arg.unwrap(), ["a", "b", "c"])
        self.assertIsNotNone(simple_args.input_file.unwrap())
        self.assertIsNotNone(simple_args.output_file.unwrap())
        self.assertEqual(simple_args.log_level.unwrap(), "DEBUG")
        self.assertEqual(simple_args.mode.unwrap(), "careful")
        self.assertListEqual(simple_args.enabled_features.unwrap(), ["CACHE", "RETRY"])
        self.assertTupleEqual(simple_args.tuple_features.unwrap(), ("CACHE", "LOGGING"))
        # optional_flag was SUPPRESS
        self.assertIsNone(simple_args.optional_flag.value)

    def test_literal_choices_enforced(self):
        parser = SimpleArguments.get_parser()
        with self.assertRaises(SystemExit):
            parser.parse_args(["-i", "1", "in.txt", "--tuple-features", "BAD", "LOGGING"])


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from typing import Generic, Protocol, TypeVar, runtime_checkable

from .base import BaseArguments

T = TypeVar("T", covariant=True)
S = TypeVar("S", bound=BaseArguments)


@runtime_checkable
class Runnable(Protocol, Generic[T]):
    def run(self) -> T: ...


class RunnableArguments(BaseArguments, ABC, Runnable[T]):
    @abstractmethod
    def run(self) -> T: ...


class SubcommandArguments(BaseArguments):
    def execute(self) -> None:
        if isinstance(last_subcommand := self.last_subcommand, Runnable):
            last_subcommand.run()
        else:
            # If there's a last_subcommand but it's not runnable, show its help
            # Otherwise, show this class's help
            if last_subcommand is not None:
                last_subcommand.get_parser().print_help()
            else:
                self.get_parser().print_help()



import unittest

from spargear import ArgumentSpec, BaseArguments


class TestGetSetAttr(unittest.TestCase):
    def test_set_get_attr(self):
        """Test setting and getting attributes in BaseArguments."""

        class ExampleArgs(BaseArguments):
            attr1: str = "default"
            attr2: int = 42
            attr3: ArgumentSpec[str] = ArgumentSpec(name_or_flags=["--attr3"], default="spec_default")
            attr4: ArgumentSpec[int] = ArgumentSpec(name_or_flags=["--attr4"], default_factory=lambda: 100)

        args = ExampleArgs()

        # Set attributes
        args.attr1 = "new_value"
        args.attr2 = 100
        args.attr3.value = "new_spec_value"
        args.attr4.value = 200

        # Get attributes
        self.assertEqual(args.attr1, "new_value")
        self.assertEqual(args.attr2, 100)
        self.assertEqual(args.attr3.value, "new_spec_value")
        self.assertEqual(args.attr4.value, 200)



import argparse
import inspect
import json
import logging
import pickle
from copy import deepcopy
from dataclasses import field, make_dataclass
from pathlib import Path
from traceback import print_exc
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from ._typing import (
    Action,
    Annotated,
    assert_type,
    extract_attr_docstrings,
    get_args,
    get_arguments_of_container_types,
    get_origin,
    get_type_hints,
    get_type_of_element_of_container_types,
    is_optional,
    sanitize_flag,
)
from .argspec import ArgumentSpec, ArgumentSpecType, ensure_no_optional
from .subcommand import SubcommandSpec

logger = logging.getLogger(__name__)


class BaseArguments:
    """Base class for defining arguments declaratively using ArgumentSpec."""

    __arguments__: Dict[str, Tuple[ArgumentSpec[object], ArgumentSpecType]]
    __subcommands__: Dict[str, SubcommandSpec["BaseArguments"]]
    __subcommand: Optional["BaseArguments"] = None

    @property
    def last_subcommand(self) -> Optional["BaseArguments"]:
        return self.__subcommand

    def __init__(self, args: Optional[Sequence[str]] = None, _internal_init: bool = False) -> None:
        """
        Initializes the BaseArguments instance and loads arguments from the command line or a given list of arguments.
        If no arguments are provided, it uses sys.argv[1:] by default.
        """
        # Initialize instance-specific argument values and specs
        self.__instance_values__: Dict[str, object] = {}
        self.__instance_specs__: Dict[str, ArgumentSpec[object]] = {}

        # only load at root (내부 생성이 아닌 경우)
        if not _internal_init:
            cls = self.__class__
            parser = cls.get_parser()
            try:
                parsed_args = parser.parse_args(args)
            except SystemExit:
                raise

            # load this class's own specs
            self.__load_from_namespace(parsed_args)

            # now walk down through any subcommands
            current_cls = cls
            current_inst: Optional["BaseArguments"] = None
            depth = 0
            while current_cls._has_subcommands():
                # 각 레벨에 맞는 dest 이름 사용
                if depth == 0:
                    dest_name = "subcommand"
                else:
                    dest_name = f"subcommand_depth_{depth}"

                subname = getattr(parsed_args, dest_name, None)
                if not subname:
                    break

                subc = current_cls.__subcommands__.get(subname)
                if not subc:
                    break

                try:
                    argument_class = subc.get_argument_class()
                except Exception:
                    break

                # Create subcommand instance with internal flag
                inst = argument_class(args=None, _internal_init=True)
                # Load values from parsed args
                inst.__load_from_namespace(parsed_args)
                current_inst = inst
                current_cls = argument_class
                depth += 1
            self.__subcommand = current_inst

    def __str__(self) -> str:
        """String representation of the BaseArguments instance."""
        return f"{self.__class__.__name__}({self.to_json(indent=2)})"

    def __getitem__(self, key: str) -> Optional[object]:
        return self.__instance_values__.get(key, self.__class__.__arguments__[key][0].value)

    def __setattr__(self, name: str, value: object) -> None:
        """Override attribute setting to store values in instance-specific storage."""
        if name not in self.__arguments__:
            # If the attribute is not an argument, use normal setattr
            super().__setattr__(name, value)
            return

        spec, spec_type = self.__arguments__[name]
        if spec_type.is_specless_type:
            # If it's a specless type, store the value directly
            self.__instance_values__[name] = value
            return

        # For ArgumentSpec types, assign the value to the instance-specific spec
        instance_specs = self.__instance_specs__
        if name in self.__instance_specs__:
            # If the spec exists in instance_specs, update its value
            instance_specs[name].value = value
        else:
            # Create a new instance copy of the spec if it doesn't exist
            instance_spec: ArgumentSpec[object] = deepcopy(spec)
            instance_spec.value = value
            instance_specs[name] = instance_spec

    def __getattribute__(self, name: str) -> object:
        """Override attribute access to return instance-specific ArgumentSpec objects or values."""
        # For special attributes, use normal access
        if TYPE_CHECKING:
            arguments = self.__arguments__
        else:
            arguments = super().__getattribute__("__arguments__")

        if name not in arguments:
            return super().__getattribute__(name)

        if TYPE_CHECKING:
            instance_values = self.__instance_values__
            instance_specs = self.__instance_specs__
        else:
            instance_values = super().__getattribute__("__instance_values__")
            instance_specs = super().__getattribute__("__instance_specs__")

        spec, spec_type = arguments[name]

        # For specless types, return the actual value if it exists
        if spec_type.is_specless_type:
            if name in instance_values:
                # If the value is already set in instance_values, return it
                return instance_values[name]
            return spec.default

        # For ArgumentSpec types, return the instance-specific spec
        if name in instance_specs:
            # If the spec exists in instance_specs, return it
            return instance_specs[name]

        # Create a new instance copy of the spec if it doesn't exist
        instance_spec: ArgumentSpec[object] = deepcopy(spec)
        instance_specs[name] = instance_spec
        return instance_spec

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        cls.__arguments__ = {}
        cls.__subcommands__ = {}

        for current_cls in reversed(cls.__mro__):
            if current_cls in (object, BaseArguments):
                continue
            # Subcommands
            for attr_value in vars(current_cls).values():
                if isinstance(attr_value, SubcommandSpec):
                    attr_value = cast(SubcommandSpec["BaseArguments"], attr_value)
                    cls.__subcommands__[attr_value.name] = attr_value

            # ArgumentSpecs
            docstrings: Dict[str, str] = extract_attr_docstrings(current_cls)
            for attr_name, attr_hint in _get_type_hints(current_cls):
                attr_value: Optional[object] = getattr(current_cls, attr_name, None)
                if isinstance(attr_value, ArgumentSpec):
                    spec: ArgumentSpec[object] = cast(ArgumentSpec[object], attr_value)
                else:
                    spec, attr_hint = _infer_spec_and_correct_typehint_from_nonspec_typehint(
                        attr_name=attr_name,
                        type_no_spec=attr_hint,
                        attr_value=attr_value,
                        docstrings=docstrings,
                    )

                if attr_name in cls.__arguments__:
                    logger.debug(f"Duplicate argument name '{attr_name}' in {current_cls.__name__}.")

                try:
                    # Extract type information from type hint
                    spec_type: ArgumentSpecType = ArgumentSpecType.from_type_hint(attr_hint)

                    # Set `choices` and `type`
                    if literals := spec_type.choices:
                        spec.choices = literals
                    if spec.type is None and (th := spec_type.type):
                        spec.type = th

                    # Determine `nargs` depending on list/tuple type
                    if tn := spec_type.tuple_nargs:
                        spec.nargs = tn
                    elif spec.nargs is None and spec_type.should_return_as_list or spec_type.should_return_as_tuple:
                        spec.nargs = "*"

                    cls.__arguments__[attr_name] = (spec, spec_type)
                except Exception as e:
                    print_exc()
                    logger.warning(f"Error processing {attr_name} in {current_cls.__name__}: {e}")
                    continue

    def get(self, key: str) -> Optional[object]:
        return self.__instance_values__.get(key, self.__class__.__arguments__[key][0].value)

    def keys(self) -> Iterable[str]:
        yield from (k for k, _v in self.items())

    def values(self) -> Iterable[object]:
        yield from (v for _k, v in self.items())

    def items(self) -> Iterable[Tuple[str, object]]:
        for key, spec, _ in self.__class__.__iter_arguments():
            value = self.__instance_values__.get(key, spec.value)
            if value is not None:
                yield key, value

    def to_dataclass(self, class_name: Optional[str] = None) -> Any:
        """Convert the BaseArguments instance to a dataclass instance.

        Args:
            class_name: Name for the generated dataclass. Defaults to {ClassName}Config.

        Returns:
            A dataclass instance with all the argument values.
        """
        if class_name is None:
            class_name = f"{self.__class__.__name__}Config"

        # Collect all argument values
        field_definitions: List[Tuple[str, type, Any]] = []
        field_values: Dict[str, Any] = {}

        for key, spec, spec_type in self.__class__.__iter_arguments():
            # Get the value from instance
            if hasattr(self, "__instance_values__") and key in self.__instance_values__:
                value = self.__instance_values__[key]
            else:
                value = spec.default

            # Determine the field type
            field_type = spec_type.type_no_optional_or_spec if spec_type.type_no_optional_or_spec is not object else Any
            if not isinstance(field_type, type):
                field_type = Any

            # Create field definition - handle mutable defaults
            if isinstance(value, (list, dict, set)):
                # Use default_factory for mutable types
                def make_factory(val: Any) -> Callable[[], Any]:
                    if hasattr(val, "copy"):
                        return lambda: val.copy()
                    else:
                        return lambda: list(val)

                field_definitions.append((
                    key,
                    cast(type, field_type),
                    field(default_factory=make_factory(value)),
                ))
            else:
                field_definitions.append((
                    key,
                    cast(type, field_type),
                    field(default=value),
                ))
            field_values[key] = value

        # Create the dataclass
        DataclassType = make_dataclass(class_name, field_definitions)

        # Create instance with current values
        return DataclassType(**field_values)

    def to_dict(self) -> Dict[str, object]:
        """Convert the BaseArguments instance to a dictionary.

        Returns:
            A dictionary with all argument names and their values.
        """
        result: Dict[str, object] = {}
        for key, spec, _ in self.__class__.__iter_arguments():
            if hasattr(self, "__instance_values__") and key in self.__instance_values__:
                value = self.__instance_values__[key]
            else:
                value = spec.default

            # Only include non-None values
            if value is not None:
                result[key] = value

        return result

    def to_json(
        self,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: Optional[Type[json.JSONEncoder]] = None,
        indent: Optional[Union[int, str]] = None,
        separators: Optional[Tuple[str, str]] = None,
        default: Optional[Callable[[object], object]] = None,
        sort_keys: bool = False,
        **kwds: Any,
    ) -> str:
        """Serialize the BaseArguments instance to JSON."""
        data = self.to_dict()

        # Convert non-serializable types
        def default_fallback(obj: object) -> object:
            if isinstance(obj, Path):
                return str(obj)
            elif hasattr(obj, "__dict__"):
                return obj.__dict__
            else:
                return str(obj)

        json_str = json.dumps(
            data,
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            cls=cls,
            indent=indent,
            separators=separators,
            default=default or default_fallback,
            sort_keys=sort_keys,
            **kwds,
        )

        return json_str

    @overload
    def to_pickle(
        self,
        protocol: Optional[int] = None,
        *,
        fix_imports: bool = True,
        buffer_callback: Optional[Callable[[object], None]] = None,
        pickler: pickle.Pickler,
    ) -> None: ...
    @overload
    def to_pickle(
        self,
        protocol: Optional[int] = None,
        *,
        fix_imports: bool = True,
        buffer_callback: Optional[Callable[[object], None]] = None,
        pickler: None = None,
    ) -> bytes: ...
    def to_pickle(
        self,
        protocol: Optional[int] = None,
        *,
        fix_imports: bool = True,
        buffer_callback: Optional[Callable[[object], None]] = None,
        pickler: Optional[pickle.Pickler] = None,
    ) -> Optional[bytes]:
        """Serialize the BaseArguments instance to a pickle file."""

        # Remove all lambda-based default factories
        for key, spec in self.__instance_specs__.items():
            if spec.default_factory is not None and callable(spec.default_factory):
                logger.warning(f"Removing default_factory from {spec.name_or_flags} in {self.__class__.__name__}.")
                spec.default_factory = None
                spec_and_spectype = self.__arguments__.get(key)
                if spec_and_spectype is None:
                    continue
                spec, _ = spec_and_spectype
                if spec.default_factory is not None and callable(spec.default_factory):
                    spec.default_factory = None

        # If a pickler is provided, use it to serialize
        if pickler is None:
            # If no pickler is provided, use default pickle serialization
            return pickle.dumps(self, protocol=protocol, fix_imports=fix_imports, buffer_callback=buffer_callback)

        # If a pickler is provided, use it to serialize
        # This allows for custom pickling behavior if needed
        pickler.dump(self)
        return None

    @classmethod
    def from_dict(cls, data: Dict[str, Any], args: Optional[Sequence[str]] = None):
        """Create a BaseArguments instance from a dictionary.

        Args:
            data: Dictionary with argument names and values.
            args: Optional command line arguments to parse first.

        Returns:
            A new BaseArguments instance with values from the dictionary.
        """
        # Create instance with command line args first (if provided)
        instance = cls(args or [])

        # Apply dictionary values only for keys not set by command line
        if hasattr(instance, "__instance_values__"):
            for key, value in data.items():
                if key in cls.__arguments__:
                    # Only set if not already set by command line args or if it's still the default
                    spec, _ = cls.__arguments__[key]

                    # Set from dict if: not in instance values, is None, or is still the default value
                    if (
                        key not in instance.__instance_values__
                        or instance.__instance_values__[key] is None
                        or instance.__instance_values__[key] == spec.default
                    ):
                        instance.__instance_values__[key] = value
                        # Also update instance specs if they exist
                        if hasattr(instance, "__instance_specs__") and key in instance.__instance_specs__:
                            instance.__instance_specs__[key].value = value

        return instance

    @classmethod
    def from_json(cls, json_data: Union[str, Path], args: Optional[Sequence[str]] = None):
        """Create a BaseArguments instance from JSON data.

        Args:
            json_data: JSON string or path to JSON file.
            args: Optional command line arguments to parse first.

        Returns:
            A new BaseArguments instance with values from the JSON.
        """
        if isinstance(json_data, Path):
            # It's a file path
            data = json.loads(Path(json_data).read_text(encoding="utf-8"))
        else:
            # It's a JSON string
            data = json.loads(str(json_data))

        return cls.from_dict(data, args)

    @classmethod
    def from_pickle(cls, path_or_bytes: Union[str, Path, bytes], args: Optional[Sequence[str]] = None):
        """Create a BaseArguments instance from a pickle file.

        Args:
            path_or_bytes: Path to the pickle file or bytes data.
            args: Optional command line arguments to parse first.

        Returns:
            A new BaseArguments instance with values from the pickle file.
        """
        if isinstance(path_or_bytes, bytes):
            # If file_path is bytes, assume it's a serialized BaseArguments instance
            data = pickle.loads(path_or_bytes)

        else:
            with open(path_or_bytes, "rb") as f:
                data = pickle.load(f)
        if isinstance(data, cls):
            return data
        else:
            raise ValueError("Bytes data does not represent a BaseArguments instance.")

    def update_from_dict(self, data: Dict[str, Any]) -> None:
        """Update the current instance with values from a dictionary.

        Args:
            data: Dictionary with argument names and values to update.
        """
        if hasattr(self, "__instance_values__"):
            for key, value in data.items():
                if key in self.__class__.__arguments__:
                    self.__instance_values__[key] = value
                    # Also update instance specs if they exist
                    if hasattr(self, "__instance_specs__") and key in self.__instance_specs__:
                        self.__instance_specs__[key].value = value

    @classmethod
    def load_config(
        cls,
        file_path: Union[str, Path],
        format: Optional[Literal["json", "pickle"]] = None,
        args: Optional[Sequence[str]] = None,
    ):
        """Load configuration from a file.

        Args:
            file_path: Path to the configuration file.
            format: File format. If None, inferred from file extension.
            args: Optional command line arguments to parse first.

        Returns:
            A new BaseArguments instance with values from the file.
        """
        path = Path(file_path)

        if format is None:
            # Infer format from extension
            if path.suffix.lower() == ".json":
                format = "json"
            elif path.suffix.lower() in (".pkl", ".pickle"):
                format = "pickle"
            else:
                raise ValueError(f"Cannot infer format from extension: {path.suffix}")

        if format == "json":
            return cls.from_json(path, args)
        elif format == "pickle":
            return cls.from_pickle(path, args)
        else:
            raise ValueError(f"Unsupported format: {format}")

    @classmethod
    def get_parser(cls) -> argparse.ArgumentParser:
        arg_parser = argparse.ArgumentParser(
            description=cls.__doc__,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            add_help=False,
        )
        arg_parser.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit.",
        )
        cls.__configure_parser(arg_parser)
        return arg_parser

    @classmethod
    def __iter_arguments(
        cls,
    ) -> Iterable[Tuple[str, ArgumentSpec[object], ArgumentSpecType]]:
        yield from ((key, spec, spec_type) for key, (spec, spec_type) in cls.__arguments__.items())

    @classmethod
    def __iter_subcommands(
        cls,
    ) -> Iterable[Tuple[str, SubcommandSpec["BaseArguments"]]]:
        yield from cls.__subcommands__.items()

    @classmethod
    def _has_subcommands(cls) -> bool:
        return bool(cls.__subcommands__)

    @classmethod
    def __add_argument_to_parser(
        cls, parser: argparse.ArgumentParser, name_or_flags: List[str], **kwargs: object
    ) -> None:
        parser.add_argument(*name_or_flags, **kwargs)  # type: ignore

    @classmethod
    def __configure_parser(cls, parser: argparse.ArgumentParser, _depth: int = 0) -> None:
        # 1) add this class's own arguments
        for key, spec, _ in cls.__iter_arguments():
            kwargs = spec.get_add_argument_kwargs()
            is_positional = not any(name.startswith("-") for name in spec.name_or_flags)
            if is_positional:
                kwargs.pop("required", None)
                cls.__add_argument_to_parser(parser, spec.name_or_flags, **kwargs)
            else:
                kwargs.setdefault("dest", key)
                cls.__add_argument_to_parser(parser, spec.name_or_flags, **kwargs)

        # 2) if there are subcommands, add them at this level
        if cls._has_subcommands():
            # 각 레벨에 고유한 dest 이름 사용
            if _depth == 0:
                dest_name = "subcommand"
            else:
                dest_name = f"subcommand_depth_{_depth}"

            subparsers = parser.add_subparsers(
                title="subcommands",
                dest=dest_name,
                metavar="subcommand",  # Always show 'subcommand' in help text
                help="Available subcommands",
                required=not cls.__arguments__ and bool(cls.__subcommands__),
            )
            for name, subc in cls.__iter_subcommands():
                subparser = subparsers.add_parser(
                    name,
                    help=subc.help,
                    description=subc.description or subc.help,
                )
                try:
                    argument_class = subc.get_argument_class()
                    argument_class.__configure_parser(subparser, _depth + 1)
                except Exception:
                    # If getting the argument class fails, skip this subcommand configuration
                    pass

    def __load_from_namespace(self, args: argparse.Namespace) -> None:
        # First, create instance-specific copies of all ArgumentSpecs
        for key, spec, spec_type in self.__class__.__iter_arguments():
            # Create a copy of the spec for this instance

            instance_spec = deepcopy(spec)
            self.__instance_specs__[key] = instance_spec

        for key, spec, spec_type in self.__class__.__iter_arguments():
            instance_spec: ArgumentSpec[object] = self.__instance_specs__[key]
            is_positional: bool = not any(n.startswith("-") for n in spec.name_or_flags)
            attr = spec.name_or_flags[0] if is_positional else (spec.dest or key)
            if not hasattr(args, attr):
                continue
            val: object = cast(object, getattr(args, attr))
            if val is argparse.SUPPRESS:
                continue

            checkable_type: Optional[type] = (
                spec.type if spec.type is not None and isinstance(spec.type, type) else None
            )

            # Type check for list/tuple
            if spec_type.should_return_as_list:
                if isinstance(val, list):
                    val = cast(List[object], val)

                elif val is not None:
                    val = [val]
                if val is not None and checkable_type is not None:
                    for v in val:
                        assert_type(v, checkable_type)

            elif spec_type.should_return_as_tuple:
                if isinstance(val, tuple):
                    val = cast(Tuple[object, ...], val)
                elif val is not None:
                    if isinstance(val, list):
                        val = tuple(cast(List[object], val))
                    else:
                        val = (val,)
                if val is not None and checkable_type is not None:
                    for v in val:
                        assert_type(v, checkable_type)

            elif val is not None and checkable_type is not None:
                assert_type(val, checkable_type)

            # Store value in instance-specific storage
            self.__instance_values__[key] = val
            # Update the instance-specific spec
            instance_spec.value = val

        # Apply default factories after all values are loaded
        for key, spec, spec_type in self.__class__.__iter_arguments():
            instance_spec = self.__instance_specs__[key]
            # Only apply default factory if no value was set from command line
            if key not in self.__instance_values__ or self.__instance_values__[key] is None:
                if instance_spec.default_factory is not None:
                    factory_value = instance_spec.default_factory()
                    self.__instance_values__[key] = factory_value
                    # Update the instance-specific spec
                    instance_spec.value = factory_value


ignored_annotations = tuple(get_type_hints(BaseArguments).keys())


def _get_type_hints(obj: type) -> Iterator[Tuple[str, object]]:
    """Get type hints for an object, excluding those in BaseArguments."""
    for k, v in get_type_hints(obj, include_extras=True).items():
        if k not in ignored_annotations:
            yield k, v


def _infer_spec_and_correct_typehint_from_nonspec_typehint(
    attr_name: str, type_no_spec: object, attr_value: object, docstrings: Dict[str, str]
) -> Tuple[ArgumentSpec[object], object]:
    action: Optional[Action] = None
    type: Optional[Callable[[str], object]] = None

    optional: bool = is_optional(type_no_spec)
    type_no_optional_or_spec: object = ensure_no_optional(type_no_spec)

    if get_origin(type_no_optional_or_spec) is Annotated:
        args = get_args(type_no_optional_or_spec)

        optional = optional or is_optional(args[0])
        type_no_optional_or_spec = ensure_no_optional(args[0])
        callable_metadata = next((a for a in args[1:] if callable(a)), None)
    else:
        callable_metadata: Optional[Callable[[str], object]] = None

    if callable_metadata is not None:
        type = callable_metadata
        if get_arguments_of_container_types(
            type_no_optional_or_spec=type_no_optional_or_spec,
            container_types=(list, tuple),
        ):
            type_no_optional_or_spec = get_type_of_element_of_container_types(
                type_no_optional_or_spec=ensure_no_optional(type_no_optional_or_spec),
                container_types=(list, tuple),
            )
    elif type_no_optional_or_spec is bool:
        if attr_value is False:
            # If the default is False, we want to set action to store_true
            action = "store_true"
        elif attr_value is True:
            # If the default is True, we want to set action to store_false
            action = "store_false"
        else:
            # If the default is None, we want to get explicit boolean value
            def get_boolean(x: str) -> bool:
                if x.lower() in ("true", "1", "yes"):
                    return True
                elif x.lower() in ("false", "0", "no"):
                    return False
                raise argparse.ArgumentTypeError(f"Invalid boolean value: {x}")

            type = get_boolean

    # Check if attr_value is a callable (potential default factory)
    default_factory: Optional[Callable[[], object]] = None
    default_value: Optional[object] = attr_value

    # If attr_value is callable and not a type, treat it as default_factory
    if callable(attr_value) and not inspect.isclass(attr_value) and attr_value is not type:
        default_factory = attr_value
        default_value = None

    spec = ArgumentSpec(
        name_or_flags=[sanitize_flag(attr_name)],
        default=default_value,
        default_factory=default_factory,
        required=default_value is None and default_factory is None and not optional,
        help=docstrings.get(attr_name, ""),
        action=action,
        type=type,
    )
    return spec, type_no_optional_or_spec



#!/usr/bin/env python3
"""Test script to verify different instances get different default factory values."""

import datetime
import unittest
import uuid
from typing import Union, Callable
from spargear.base import BaseArguments, ArgumentSpec


class TestArguments(BaseArguments):
    """Test arguments with default factory support."""

    # Default factory using lambda
    session_id: Union[str, Callable[[], str]] = lambda: str(uuid.uuid4())
    """Unique session ID (generated at parse time)"""

    # Explicit ArgumentSpec with default_factory
    timestamp: ArgumentSpec[str] = ArgumentSpec(
        name_or_flags=["--timestamp"],
        default_factory=lambda: datetime.datetime.now().isoformat(),
        help="Current timestamp (auto-generated)",
    )


class TestDifferentInstances(unittest.TestCase):
    def test_different_instances(self) -> None:
        """Test that different instances get different default factory values."""
        print("Testing different instances...")

        # Create multiple instances
        args1 = TestArguments([])
        args2 = TestArguments([])
        args3 = TestArguments([])

        print(f"Instance 1 session_id: {args1.get('session_id')}")
        print(f"Instance 1 timestamp: {args1.get('timestamp')}")

        print(f"Instance 2 session_id: {args2.get('session_id')}")
        print(f"Instance 2 timestamp: {args2.get('timestamp')}")

        print(f"Instance 3 session_id: {args3.get('session_id')}")
        print(f"Instance 3 timestamp: {args3.get('timestamp')}")

        # Check if values are different
        session_ids = [
            args1.get("session_id"),
            args2.get("session_id"),
            args3.get("session_id"),
        ]
        timestamps = [
            args1.get("timestamp"),
            args2.get("timestamp"),
            args3.get("timestamp"),
        ]

        print(
            f"\nAll session IDs are unique: {len(set(session_ids)) == len(session_ids)}"
        )
        print(f"All timestamps are unique: {len(set(timestamps)) == len(timestamps)}")


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import List, Optional, Tuple

from spargear import Annotated, ArgumentSpec, BaseArguments


class SpeclessBasicArguments(BaseArguments):
    """Example argument parser demonstrating specless type."""

    optional_int_with_default: Optional[int] = None
    optional_int_without_default: Optional[int]
    float_with_default: float = 0.0
    """float_with_default"""
    float_without_default: float
    float_with_str_default: float = "1.23"  # type: ignore[assignment]
    list_of_ints_without_default: List[int]
    """list_of_ints_without_default"""
    some_argument: ArgumentSpec[bool] = ArgumentSpec(["--some-argument"], action="store_true", help="some_argument")
    """This should be ignored"""


class TestSpeclessBasic(unittest.TestCase):
    def test_specless_type_no_optional(self):
        args = SpeclessBasicArguments([
            "--float-without-default",
            "3.14",
            "--list-of-ints-without-default",
            "1",
            "2",
            "3",
        ])
        self.assertEqual(args.optional_int_with_default, None)
        self.assertEqual(args.optional_int_without_default, None)
        self.assertEqual(args.float_with_default, 0.0)
        self.assertEqual(args.float_without_default, 3.14)
        self.assertEqual(args.float_with_str_default, 1.23)
        self.assertEqual(args.list_of_ints_without_default, [1, 2, 3])

    def test_specless_type_optional(self):
        args = SpeclessBasicArguments([
            "--optional-int-with-default",
            "3",
            "--optional-int-without-default",
            "4",
            "--float-without-default",
            "3.14",
            "--list-of-ints-without-default",
            "1",
            "2",
            "3",
        ])
        self.assertEqual(args.optional_int_with_default, 3)
        self.assertEqual(args.optional_int_without_default, 4)

    def test_specless_type_docs(self):
        args = SpeclessBasicArguments([
            "--float-without-default",
            "3.14",
            "--list-of-ints-without-default",
            "1",
            "2",
            "3",
        ])
        self.assertEqual(args.__arguments__["float_with_default"][0].help, "float_with_default")
        self.assertEqual(args.__arguments__["list_of_ints_without_default"][0].help, "list_of_ints_without_default")
        self.assertEqual(args.__arguments__["some_argument"][0].help, "some_argument")


class SpeclessBooleanArguments(BaseArguments):
    """Example argument parser demonstrating specless boolean arguments."""

    bool_without_default: bool
    """bool_without_default"""
    optional_bool_with_default: Optional[bool] = None
    """optional_bool_with_default"""
    optional_bool_without_default: Optional[bool]
    """optional_bool_without_default"""
    bool_with_default_false: bool = False
    """bool_with_default_false"""
    bool_with_default_true: bool = True
    """bool_with_default_true"""


class SpeclessPositionalArguments(BaseArguments):
    flag: int
    """--flag"""
    POSITIONAL: int
    """positional"""
    AnotherFlag: int


class TestSpeclessBooleanArguments(unittest.TestCase):
    def test_basic_things(self):
        with self.assertRaises(SystemExit):
            SpeclessBooleanArguments([])

        args = SpeclessBooleanArguments(["--bool-without-default", "False"])
        self.assertEqual(args.bool_without_default, False)
        self.assertEqual(args.optional_bool_with_default, None)
        self.assertEqual(args.optional_bool_without_default, None)
        self.assertEqual(args.bool_with_default_false, False)
        self.assertEqual(args.bool_with_default_true, True)

    def test_explicit_bool(self):
        args = SpeclessBooleanArguments(["--bool-without-default", "False"])
        self.assertEqual(args.bool_without_default, False)

        args = SpeclessBooleanArguments(["--bool-without-default", "True"])
        self.assertEqual(args.bool_without_default, True)

        with self.assertRaises(SystemExit):
            SpeclessBooleanArguments(["--bool-without-default", "not-a-bool"])

    def test_store_action(self):
        args = SpeclessBooleanArguments(["--bool-without-default", "False"])
        self.assertEqual(args.bool_with_default_false, False)
        self.assertEqual(args.bool_with_default_true, True)

        args = SpeclessBooleanArguments([
            "--bool-without-default",
            "False",
            "--bool-with-default-true",
            "--bool-with-default-false",
        ])
        self.assertEqual(args.bool_with_default_false, True)
        self.assertEqual(args.bool_with_default_true, False)


class TestSpeclessTupleAndLists(unittest.TestCase):
    def test_basic_things(self):
        class SpeclessTupleAndLists(BaseArguments):
            """Example argument parser demonstrating specless tuple and list arguments."""

            tuple_of_ints: Tuple[int, int]
            """tuple_of_ints"""
            list_of_ints_without_default: List[int]
            """list_of_ints_without_default"""

        with self.assertRaises(SystemExit):
            SpeclessTupleAndLists([])
        with self.assertRaises(SystemExit):
            SpeclessTupleAndLists(["--tuple-of-ints", "1"])
        with self.assertRaises(SystemExit):
            SpeclessTupleAndLists(["--list-of-ints", "1", "2"])
        args1 = SpeclessTupleAndLists(["--tuple-of-ints", "1", "2", "--list-of-ints", "3"])
        self.assertEqual(args1.tuple_of_ints, (1, 2))
        self.assertEqual(args1.list_of_ints_without_default, [3])
        args1 = SpeclessTupleAndLists(["--tuple-of-ints", "1", "2", "--list-of-ints"])
        self.assertEqual(args1.tuple_of_ints, (1, 2))
        self.assertEqual(args1.list_of_ints_without_default, [])

    def test_basic_things2(self):
        class SpeclessTupleAndLists(BaseArguments):
            list_of_ints_with_default: List[int] = [1, 2, 3]
            """list_of_ints_with_default"""

        args2 = SpeclessTupleAndLists([])
        self.assertEqual(args2.list_of_ints_with_default, [1, 2, 3])
        args2 = SpeclessTupleAndLists(["--list-of-ints-with-default", "4", "5"])
        self.assertEqual(args2.list_of_ints_with_default, [4, 5])

    def test_basic_things3(self):
        class SpeclessTupleAndLists(BaseArguments):
            optional_list_of_ints_with_default: Optional[List[int]] = None
            """optional_list_of_ints_with_default"""

        args3 = SpeclessTupleAndLists([])
        self.assertEqual(args3.optional_list_of_ints_with_default, None)
        args3 = SpeclessTupleAndLists(["--optional-list-of-ints-with-default", "4", "5"])
        self.assertEqual(args3.optional_list_of_ints_with_default, [4, 5])

    def test_basic_things4(self):
        class SpeclessTupleAndLists(BaseArguments):
            optional_list_of_ints_without_default: Optional[List[int]]
            """optional_list_of_ints_without_default"""

        args4 = SpeclessTupleAndLists([])
        self.assertEqual(args4.optional_list_of_ints_without_default, None)
        args4 = SpeclessTupleAndLists(["--optional-list-of-ints-without-default", "4", "5"])
        self.assertEqual(args4.optional_list_of_ints_without_default, [4, 5])


class TestSpeclessPositionalArguments(unittest.TestCase):
    def test_basic_things(self):
        args = SpeclessPositionalArguments(["2", "--flag", "1", "--anotherflag", "3"])
        self.assertEqual(args.flag, 1)
        self.assertEqual(args.POSITIONAL, 2)
        self.assertEqual(args.AnotherFlag, 3)


class TestSpeclessInheritance(unittest.TestCase):
    def test_basic_things(self):
        class SpeclessInheritance(BaseArguments):
            """Example argument parser demonstrating specless inheritance."""

            flag1: int
            """--flag1"""

        class SpeclessInheritance2(SpeclessInheritance):
            flag1: str  # pyright: ignore[reportIncompatibleVariableOverride]
            """--flag1"""
            flag2: int
            """--flag2"""

        args = SpeclessInheritance2(["--flag1", "1", "--flag2", "2"])
        self.assertEqual(args.flag1, "1")
        self.assertEqual(args.flag2, 2)


class TestAnnotated(unittest.TestCase):
    def test_basic_things(self):
        def negation(x: str) -> int:
            return -int(x)

        class SpeclessAnnotated(BaseArguments):
            a: Annotated[int, negation]
            """--a"""
            b: int
            """--b"""
            c: Annotated[int, None]

        args = SpeclessAnnotated(["--a", "1", "--b", "2", "--c", "3"])
        self.assertEqual(args.a, -1)
        self.assertEqual(args.b, 2)
        self.assertEqual(args.c, 3)

    def test_list(self):
        def as_list(x: str) -> List[int]:
            return [int(i) for i in x.split(",")]

        class SpeclessAnnotated(BaseArguments):
            a: Annotated[List[int], None]
            b: Annotated[List[int], None] = [-2, -3, -4]
            c: Annotated[List[int], as_list]
            d: Annotated[List[int], as_list] = [-4, -5, -6]
            e: Annotated[int, as_list]  # Wrong case! (anti-pattern)
            f: Annotated[int, as_list] = 1  # Wrong case! (anti-pattern)

        args = SpeclessAnnotated([
            "--a",
            "1",
            "2",
            "3",
            "--b",
            "2",
            "3",
            "4",
            "--c",
            "3,4,5",
            "--d",
            "4,5,6",
            "--e",
            "5",
            "--f",
            "6",
        ])
        self.assertEqual(args.a, [1, 2, 3])
        self.assertEqual(args.b, [2, 3, 4])
        self.assertEqual(args.c, [3, 4, 5])
        self.assertEqual(args.d, [4, 5, 6])
        self.assertEqual(args.e, [5])
        self.assertEqual(args.f, [6])


if __name__ == "__main__":
    unittest.main()
